import sys
import tempfile
from abc import ABC, abstractmethod
from collections.abc import Generator
from contextlib import contextmanager
from unittest.mock import patch

import dagster._check as check
import pytest
from dagster import file_relative_path
from dagster._core.instance import DagsterInstance, InstanceType
from dagster._core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
from dagster._core.run_coordinator import DefaultRunCoordinator
from dagster._core.storage.event_log.sqlite import ConsolidatedSqliteEventLogStorage
from dagster._core.storage.local_compute_log_manager import LocalComputeLogManager
from dagster._core.storage.root import LocalArtifactStorage
from dagster._core.storage.runs import InMemoryRunStorage
from dagster._core.test_utils import instance_for_test
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._core.workspace.context import WorkspaceProcessContext, WorkspaceRequestContext
from dagster._core.workspace.load_target import (
    GrpcServerTarget,
    ModuleTarget,
    PythonFileTarget,
    WorkspaceFileTarget,
)
from dagster._grpc.client import DagsterGrpcClient
from dagster._grpc.server import GrpcServerProcess, wait_for_grpc_server
from dagster._utils import safe_tempfile_path
from dagster._utils.merger import merge_dicts
from dagster._utils.test import FilesystemTestScheduler
from dagster._utils.test.postgres_instance import TestPostgresInstance
from dagster_graphql import DagsterGraphQLClient
from dagster_graphql.test.utils import execute_dagster_graphql
from dagster_shared.ipc import open_ipc_subprocess
from graphql import print_ast

try:
    from gql.client import GraphQLRequest

    HAS_GRAPHQL_REQUEST = True
except ImportError:
    # gql <3.5.0 doesn't have GraphQLRequest
    HAS_GRAPHQL_REQUEST = False


def get_main_loadable_target_origin():
    return LoadableTargetOrigin(
        executable_path=sys.executable,
        python_file=file_relative_path(__file__, "repo.py"),
        attribute="test_repo",
    )


@contextmanager
def graphql_postgres_instance(
    overrides=None, synchronous_run_launcher=False, synchronous_run_coordinator=False
):
    with tempfile.TemporaryDirectory() as temp_dir:
        with TestPostgresInstance.docker_service_up_or_skip(
            file_relative_path(__file__, "docker-compose.yml"),
            "test-postgres-db-graphql",
        ) as pg_conn_string:
            TestPostgresInstance.clean_run_storage(pg_conn_string)
            TestPostgresInstance.clean_event_log_storage(pg_conn_string)
            TestPostgresInstance.clean_schedule_storage(pg_conn_string)
            with instance_for_test(
                temp_dir=temp_dir,
                overrides=merge_dicts(
                    {
                        "run_storage": {
                            "module": "dagster_postgres.run_storage.run_storage",
                            "class": "PostgresRunStorage",
                            "config": {"postgres_url": pg_conn_string},
                        },
                        "event_log_storage": {
                            "module": "dagster_postgres.event_log.event_log",
                            "class": "PostgresEventLogStorage",
                            "config": {"postgres_url": pg_conn_string},
                        },
                        "schedule_storage": {
                            "module": "dagster_postgres.schedule_storage.schedule_storage",
                            "class": "PostgresScheduleStorage",
                            "config": {"postgres_url": pg_conn_string},
                        },
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                    },
                    overrides if overrides else {},
                ),
                synchronous_run_launcher=synchronous_run_launcher,
                synchronous_run_coordinator=synchronous_run_coordinator,
            ) as instance:
                yield instance


class MarkedManager:
    """MarkedManagers are passed to GraphQLContextVariants. They contain
    a contextmanager function "manager_fn" that yield the relevant
    instance, and it includes marks that will be applied to any
    context-variant-driven test case that includes this MarkedManager.

    See InstanceManagers for an example construction.

    See GraphQLContextVariant for further information
    """

    def __init__(self, manager_fn, marks):
        self.manager_fn = check.callable_param(manager_fn, "manager_fn")
        self.marks = check.list_param(marks, "marks")


class InstanceManagers:
    @staticmethod
    def non_launchable_sqlite_instance():
        @contextmanager
        def _non_launchable_sqlite_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_launcher": {
                            "module": "dagster._core.test_utils",
                            "class": "ExplodingRunLauncher",
                        },
                    },
                    synchronous_run_coordinator=True,
                ) as instance:
                    yield instance

        return MarkedManager(
            _non_launchable_sqlite_instance,
            [Marks.sqlite_instance, Marks.non_launchable],
        )

    @staticmethod
    def non_launchable_postgres_instance():
        @contextmanager
        def _non_launchable_postgres_instance():
            with graphql_postgres_instance(
                overrides={
                    "run_launcher": {
                        "module": "dagster._core.test_utils",
                        "class": "ExplodingRunLauncher",
                    },
                },
                synchronous_run_coordinator=True,
            ) as instance:
                yield instance

        return MarkedManager(
            _non_launchable_postgres_instance,
            [Marks.postgres_instance, Marks.non_launchable],
        )

    @staticmethod
    def sqlite_instance_with_sync_run_launcher():
        @contextmanager
        def _sqlite_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                    },
                    synchronous_run_launcher=True,
                ) as instance:
                    yield instance

        return MarkedManager(_sqlite_instance, [Marks.sqlite_instance, Marks.sync_run_launcher])

    # Runs launched with this instance won't actually execute since the graphql test suite
    # doesn't run the daemon process that launches queued runs
    @staticmethod
    def sqlite_instance_with_queued_run_coordinator():
        @contextmanager
        def _sqlite_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_coordinator": {
                            "module": "dagster._core.run_coordinator.queued_run_coordinator",
                            "class": "QueuedRunCoordinator",
                        },
                    },
                ) as instance:
                    yield instance

        return MarkedManager(
            _sqlite_instance, [Marks.sqlite_instance, Marks.queued_run_coordinator]
        )

    @staticmethod
    def sqlite_instance_with_default_run_launcher():
        @contextmanager
        def _sqlite_instance_with_default_hijack():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                    },
                    synchronous_run_coordinator=True,
                ) as instance:
                    yield instance

        return MarkedManager(
            _sqlite_instance_with_default_hijack,
            [Marks.sqlite_instance, Marks.default_run_launcher],
        )

    @staticmethod
    def postgres_instance_with_sync_run_launcher():
        @contextmanager
        def _postgres_instance():
            with graphql_postgres_instance(
                synchronous_run_launcher=True,
                synchronous_run_coordinator=True,
            ) as instance:
                yield instance

        return MarkedManager(
            _postgres_instance,
            [Marks.postgres_instance, Marks.sync_run_launcher],
        )

    @staticmethod
    def postgres_instance_with_default_run_launcher():
        @contextmanager
        def _postgres_instance_with_default_hijack():
            with graphql_postgres_instance(synchronous_run_coordinator=True) as instance:
                yield instance

        return MarkedManager(
            _postgres_instance_with_default_hijack,
            [Marks.postgres_instance, Marks.default_run_launcher],
        )

    @staticmethod
    def consolidated_sqlite_instance():
        @contextmanager
        def _sqlite_asset_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                instance = DagsterInstance(
                    instance_type=InstanceType.EPHEMERAL,
                    local_artifact_storage=LocalArtifactStorage(temp_dir),
                    run_storage=InMemoryRunStorage(),
                    event_storage=ConsolidatedSqliteEventLogStorage(temp_dir),
                    compute_log_manager=LocalComputeLogManager(temp_dir),
                    run_coordinator=DefaultRunCoordinator(),
                    run_launcher=SyncInMemoryRunLauncher(),
                    scheduler=FilesystemTestScheduler(temp_dir),
                )
                yield instance

        return MarkedManager(_sqlite_asset_instance, [Marks.asset_aware_instance])

    @staticmethod
    def default_concurrency_sqlite_instance():
        @contextmanager
        def _sqlite_with_default_concurrency_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_coordinator": {
                            "module": "dagster._core.run_coordinator.queued_run_coordinator",
                            "class": "QueuedRunCoordinator",
                        },
                        "concurrency": {
                            "default_op_concurrency_limit": 1,
                        },
                    },
                ) as instance:
                    yield instance

        return MarkedManager(
            _sqlite_with_default_concurrency_instance,
            [Marks.sqlite_instance, Marks.queued_run_coordinator],
        )


class EnvironmentManagers:
    @staticmethod
    def managed_grpc(target=None, location_name="test_location"):
        @contextmanager
        def _mgr_fn(instance, read_only):
            """Relies on webserver to load the code location in a subprocess and manage its lifecyle."""
            loadable_target_origin = (
                target if target is not None else get_main_loadable_target_origin()
            )
            with WorkspaceProcessContext(
                instance,
                (
                    PythonFileTarget(
                        python_file=loadable_target_origin.python_file,
                        attribute=loadable_target_origin.attribute,
                        working_directory=loadable_target_origin.working_directory,
                        location_name=location_name,
                    )
                    if loadable_target_origin.python_file
                    else ModuleTarget(
                        module_name=loadable_target_origin.module_name,  # pyright: ignore[reportArgumentType]
                        attribute=loadable_target_origin.attribute,
                        working_directory=loadable_target_origin.working_directory,
                        location_name=location_name,
                    )
                ),
                version="",
                read_only=read_only,
            ) as workspace_process_context:
                yield workspace_process_context

        return MarkedManager(_mgr_fn, [Marks.managed_grpc_env])

    @staticmethod
    def deployed_grpc(target=None, location_name="test_location"):
        """Launches a code server in a "dagster api grpc" subprocess."""

        @contextmanager
        def _mgr_fn(instance, read_only):
            with GrpcServerProcess(
                instance_ref=instance.get_ref(),
                location_name=location_name,
                loadable_target_origin=(
                    target if target is not None else get_main_loadable_target_origin()
                ),
                wait_on_exit=True,
            ) as server_process:
                api_client = server_process.create_client()
                with WorkspaceProcessContext(
                    instance,
                    GrpcServerTarget(
                        port=api_client.port,
                        socket=api_client.socket,
                        host=api_client.host,  # pyright: ignore[reportArgumentType]
                        location_name=location_name,
                    ),
                    version="",
                    read_only=read_only,
                ) as workspace:
                    yield workspace

        return MarkedManager(_mgr_fn, [Marks.deployed_grpc_env])

    @staticmethod
    def code_server_cli_grpc(target=None, location_name="test_location"):
        """Launches a code server in a "dagster code-server start" subprocess (which will
        in turn open up a `dagster api grpc` subprocess that actually loads the code location).
        """

        @contextmanager
        def _mgr_fn(instance, read_only):
            loadable_target_origin = target or get_main_loadable_target_origin()
            with safe_tempfile_path() as socket:
                subprocess_args = [  # pyright: ignore[reportOperatorIssue]
                    "dagster",
                    "code-server",
                    "start",
                    "--socket",
                    socket,
                ] + loadable_target_origin.get_cli_args()

                server_process = open_ipc_subprocess(subprocess_args)

                client = DagsterGrpcClient(port=None, socket=socket, host="localhost")

                try:
                    wait_for_grpc_server(server_process, client, subprocess_args)
                    with WorkspaceProcessContext(
                        instance,
                        GrpcServerTarget(
                            port=None,
                            socket=socket,
                            host="localhost",
                            location_name=location_name,
                        ),
                        version="",
                        read_only=read_only,
                    ) as workspace:
                        yield workspace
                finally:
                    client.shutdown_server()
                    server_process.wait(timeout=30)

        return MarkedManager(_mgr_fn, [Marks.code_server_cli_grpc_env])

    @staticmethod
    def multi_location():
        @contextmanager
        def _mgr_fn(instance, read_only):
            """Goes out of process but same process as host process."""
            with WorkspaceProcessContext(
                instance,
                WorkspaceFileTarget(paths=[file_relative_path(__file__, "multi_location.yaml")]),
                version="",
                read_only=read_only,
            ) as workspace:
                yield workspace

        return MarkedManager(_mgr_fn, [Marks.multi_location])

    @staticmethod
    def lazy_repository():
        @contextmanager
        def _mgr_fn(instance, read_only):
            """Goes out of process but same process as host process."""
            with WorkspaceProcessContext(
                instance,
                PythonFileTarget(
                    python_file=file_relative_path(__file__, "repo.py"),
                    attribute="test_dict_repo",
                    working_directory=None,
                    location_name="test_location",
                ),
                version="",
                read_only=read_only,
            ) as workspace:
                yield workspace

        return MarkedManager(_mgr_fn, [Marks.lazy_repository])


class Marks:
    # Instance type makes
    sqlite_instance = pytest.mark.sqlite_instance
    postgres_instance = pytest.mark.postgres_instance

    # Run launcher variants
    sync_run_launcher = pytest.mark.sync_run_launcher
    default_run_launcher = pytest.mark.default_run_launcher
    queued_run_coordinator = pytest.mark.queued_run_coordinator
    non_launchable = pytest.mark.non_launchable

    # Code location marks
    multi_location = pytest.mark.multi_location
    managed_grpc_env = pytest.mark.managed_grpc_env
    deployed_grpc_env = pytest.mark.deployed_grpc_env
    code_server_cli_grpc_env = pytest.mark.code_server_cli_grpc_env

    lazy_repository = pytest.mark.lazy_repository

    # Asset-aware sqlite variants
    asset_aware_instance = pytest.mark.asset_aware_instance

    # Backfill daemon variants
    backfill_daemon = pytest.mark.backfill_daemon

    # Readonly context variant
    read_only = pytest.mark.read_only

    # Common mark to all test suite tests
    graphql_context_test_suite = pytest.mark.graphql_context_test_suite


def none_manager():
    @contextmanager
    def _yield_none(*_args, **_kwargs):
        yield None

    return MarkedManager(_yield_none, [])


class GraphQLContextVariant:
    """An instance of this class represents a context variant that will be run
    against *every* method in the test class, defined as a class
    created by inheriting from make_graphql_context_test_suite.

    It comes with a number of static methods with prebuilt context variants.
    e.g. in_memory_in_process_start

    One can also make bespoke context variants, provided you configure it properly
    with MarkedMembers that produce its members.

    Args:
    marked_instance_mgr (MarkedManager): The manager_fn
    within it must be a contextmanager that takes zero arguments and yields
    a DagsterInstance

    See InstanceManagers for examples

    marked_environment_mgr (MarkedManager): The manager_fn with in
    must be a contextmanager takes a default ReconstructableRepo and
    yields a list of RepositoryLocation.

    See EnvironmentManagers for examples

    test_id [Optional] (str): This assigns a test_id to test parameterized with this
    variant. This is highly convenient for running a particular variant across
    the entire test suite, without running all the other variants.

    e.g.
    pytest python_modules/dagster-graphql/dagster_graphql_tests/ -s -k in_memory_in_process_start

    Will run all tests that use the in_memory_in_process_start, which will get a lot
    of code coverage while being very fast to run.

    All tests managed by this system are marked with "graphql_context_test_suite".
    """

    def __init__(self, marked_instance_mgr, marked_environment_mgr, read_only=False, test_id=None):
        self.marked_instance_mgr = check.inst_param(
            marked_instance_mgr, "marked_instance_mgr", MarkedManager
        )
        self.marked_environment_mgr = check.inst_param(
            marked_environment_mgr, "marked_environment_mgr", MarkedManager
        )
        self.read_only = check.bool_param(read_only, "read_only")
        self.test_id = check.opt_str_param(test_id, "test_id")
        self.marks = (
            marked_instance_mgr.marks
            + marked_environment_mgr.marks
            + ([Marks.read_only] if read_only else [])
        )

    @property
    def instance_mgr(self):
        return self.marked_instance_mgr.manager_fn

    @property
    def environment_mgr(self):
        return self.marked_environment_mgr.manager_fn

    @staticmethod
    def sqlite_with_queued_run_coordinator_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_queued_run_coordinator(),
            EnvironmentManagers.managed_grpc(),
            test_id="sqlite_with_queued_run_coordinator_managed_grpc_env",
        )

    @staticmethod
    def sqlite_with_default_run_launcher_managed_grpc_env(
        target=None, location_name="test_location"
    ):
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.managed_grpc(target, location_name),
            test_id="sqlite_with_default_run_launcher_managed_grpc_env",
        )

    @staticmethod
    def sqlite_read_only_with_default_run_launcher_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.managed_grpc(),
            read_only=True,
            test_id="sqlite_read_only_with_default_run_launcher_managed_grpc_env",
        )

    @staticmethod
    def sqlite_with_default_run_launcher_deployed_grpc_env(
        target=None, location_name="test_location"
    ):
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.deployed_grpc(target, location_name),
            test_id="sqlite_with_default_run_launcher_deployed_grpc_env",
        )

    @staticmethod
    def sqlite_with_default_run_launcher_code_server_cli_env(
        target=None, location_name="test_location"
    ):
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.code_server_cli_grpc(target, location_name),
            test_id="sqlite_with_default_run_launcher_code_server_cli_env",
        )

    @staticmethod
    def sqlite_with_default_concurrency_managed_grpc_env(
        target=None, location_name="test_location"
    ):
        return GraphQLContextVariant(
            InstanceManagers.default_concurrency_sqlite_instance(),
            EnvironmentManagers.managed_grpc(target, location_name),
            test_id="sqlite_with_default_concurrency_managed_grpc_env",
        )

    @staticmethod
    def postgres_with_default_run_launcher_managed_grpc_env(
        target=None, location_name="test_location"
    ):
        return GraphQLContextVariant(
            InstanceManagers.postgres_instance_with_default_run_launcher(),
            EnvironmentManagers.managed_grpc(target, location_name),
            test_id="postgres_with_default_run_launcher_managed_grpc_env",
        )

    @staticmethod
    def postgres_with_default_run_launcher_deployed_grpc_env(
        target=None, location_name="test_location"
    ):
        return GraphQLContextVariant(
            InstanceManagers.postgres_instance_with_default_run_launcher(),
            EnvironmentManagers.deployed_grpc(target, location_name),
            test_id="postgres_with_default_run_launcher_deployed_grpc_env",
        )

    @staticmethod
    def non_launchable_sqlite_instance_multi_location():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.multi_location(),
            test_id="non_launchable_sqlite_instance_multi_location",
        )

    @staticmethod
    def non_launchable_sqlite_instance_lazy_repository():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.lazy_repository(),
            test_id="non_launchable_sqlite_instance_lazy_repository",
        )

    @staticmethod
    def non_launchable_sqlite_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="non_launchable_sqlite_instance_managed_grpc_env",
        )

    @staticmethod
    def non_launchable_sqlite_instance_deployed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.deployed_grpc(),
            test_id="non_launchable_sqlite_instance_deployed_grpc_env",
        )

    @staticmethod
    def non_launchable_postgres_instance_multi_location():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_postgres_instance(),
            EnvironmentManagers.multi_location(),
            test_id="non_launchable_postgres_instance_multi_location",
        )

    @staticmethod
    def non_launchable_postgres_instance_lazy_repository():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_postgres_instance(),
            EnvironmentManagers.lazy_repository(),
            test_id="non_launchable_postgres_instance_lazy_repository",
        )

    @staticmethod
    def non_launchable_postgres_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_postgres_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="non_launchable_postgres_instance_managed_grpc_env",
        )

    @staticmethod
    def consolidated_sqlite_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.consolidated_sqlite_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="asset_aware_instance_in_process_env",
        )

    @staticmethod
    def all_variants():
        """There is a test case that keeps this up-to-date. If you add a static
        method that returns a GraphQLContextVariant you have to add it to this
        list in order for tests to pass.
        """
        return [
            GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.sqlite_read_only_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_run_launcher_deployed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_run_launcher_code_server_cli_env(),
            GraphQLContextVariant.sqlite_with_queued_run_coordinator_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_deployed_grpc_env(),
            GraphQLContextVariant.non_launchable_sqlite_instance_multi_location(),
            GraphQLContextVariant.non_launchable_sqlite_instance_managed_grpc_env(),
            GraphQLContextVariant.non_launchable_sqlite_instance_deployed_grpc_env(),
            GraphQLContextVariant.non_launchable_sqlite_instance_lazy_repository(),
            GraphQLContextVariant.non_launchable_postgres_instance_multi_location(),
            GraphQLContextVariant.non_launchable_postgres_instance_managed_grpc_env(),
            GraphQLContextVariant.non_launchable_postgres_instance_lazy_repository(),
            GraphQLContextVariant.consolidated_sqlite_instance_managed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_concurrency_managed_grpc_env(),
        ]

    @staticmethod
    def all_executing_variants(target=None, location_name="test_location"):
        return [
            GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(
                target, location_name
            ),
            GraphQLContextVariant.sqlite_with_default_run_launcher_deployed_grpc_env(
                target, location_name
            ),
            GraphQLContextVariant.sqlite_with_default_run_launcher_code_server_cli_env(
                target, location_name
            ),
            GraphQLContextVariant.postgres_with_default_run_launcher_managed_grpc_env(
                target, location_name
            ),
            GraphQLContextVariant.postgres_with_default_run_launcher_deployed_grpc_env(
                target, location_name
            ),
        ]

    @staticmethod
    def all_readonly_variants():
        """Return all read only variants. If you try to run any mutation these will error."""
        return _variants_with_mark(GraphQLContextVariant.all_variants(), pytest.mark.read_only)

    @staticmethod
    def all_non_launchable_variants():
        """Return all non_launchable variants. If you try to start or launch these will error."""
        return _variants_with_mark(GraphQLContextVariant.all_variants(), pytest.mark.non_launchable)

    @staticmethod
    def all_multi_location_variants():
        return _variants_with_mark(GraphQLContextVariant.all_variants(), pytest.mark.multi_location)


def _variants_with_mark(variants, mark):
    def _yield_all():
        for variant in variants:
            if mark in variant.marks:
                yield variant

    return list(_yield_all())


def _variants_without_marks(variants, marks):
    def _yield_all():
        for variant in variants:
            if all(mark not in variant.marks for mark in marks):
                yield variant

    return list(_yield_all())


@contextmanager
def manage_graphql_context(context_variant):
    with context_variant.instance_mgr() as instance:
        with context_variant.environment_mgr(
            instance, context_variant.read_only
        ) as workspace_process_context:
            yield workspace_process_context


class _GraphQLContextTestSuite(ABC):
    @abstractmethod
    @contextmanager
    def yield_graphql_context(
        self, class_scoped_context
    ) -> Generator[WorkspaceRequestContext, None, None]:
        pass

    @contextmanager
    def graphql_context_for_request(self, request):
        check.param_invariant(
            isinstance(request.param, GraphQLContextVariant),
            "request",
            "params in fixture must be List[GraphQLContextVariant]",
        )
        with manage_graphql_context(request.param) as graphql_context:
            yield graphql_context


def graphql_context_variants_fixture(context_variants):
    check.list_param(context_variants, "context_variants", of_type=GraphQLContextVariant)

    def _wrap(fn):
        return pytest.fixture(
            name="class_scoped_graphql_context",
            scope="class",
            params=[
                pytest.param(
                    context_variant,
                    id=context_variant.test_id,
                    marks=context_variant.marks + [Marks.graphql_context_test_suite],
                )
                for context_variant in context_variants
            ],
        )(fn)

    return _wrap


def make_graphql_context_test_suite(context_variants):
    """Arguments:
        context_variants (List[GraphQLContextVariant]): List of runs to run per test in this class.

        This is the base class factory for test suites in the dagster-graphql test.

        The goal of this suite is to make it straightforward to run tests
        against multiple graphql_contexts, have a coherent lifecycle for those
        contexts.

        GraphQLContextVariant has a number of static methods to provide common run configurations
        as well as common groups of run configuration

        One can also make bespoke GraphQLContextVariants with specific implementations
        of DagsterInstance, RepositoryLocation, and so forth. See that class
        for more details.

    Example:
    class TestAThing(
        make_graphql_context_test_suite(
            context_variants=[GraphQLContextVariant.in_memory_in_process_start()]
        )
    ):
        def test_graphql_context_exists(self, graphql_context):
            assert graphql_context
    """
    check.list_param(context_variants, "context_variants", of_type=GraphQLContextVariant)

    class _SpecificTestSuiteBase(_GraphQLContextTestSuite):
        @graphql_context_variants_fixture(context_variants=context_variants)
        def yield_class_scoped_graphql_context(self, request):
            with self.graphql_context_for_request(request) as graphql_context:
                yield graphql_context

        @pytest.fixture(name="graphql_context")
        def graphql_context_fixture(self, class_scoped_graphql_context):
            with self.yield_graphql_context(class_scoped_graphql_context) as context:
                yield context

        @pytest.fixture(name="graphql_client")
        def graphql_client_fixture(self, graphql_context):
            with self.yield_graphql_client(graphql_context) as client:
                yield client

        @contextmanager
        def yield_graphql_context(
            self, class_scoped_context
        ) -> Generator[WorkspaceRequestContext, None, None]:
            instance = class_scoped_context.instance
            instance.wipe()
            instance.wipe_all_schedules()
            with class_scoped_context.create_request_context() as request_context:
                yield request_context
            # ensure that any runs launched by the test are cleaned up
            # Since launcher is lazy loaded, we don't need to do anyting if it's None
            if instance._run_launcher:  # noqa: SLF001
                instance._run_launcher.join()  # noqa: SLF001

        @contextmanager
        def yield_graphql_client(self, context) -> Generator[DagsterGraphQLClient, None, None]:
            class MockedGraphQLClient:
                def execute(self, gql_query, variable_values=None):
                    # Handle both gql v3 (DocumentNode) and v4 (GraphQLRequest)
                    if HAS_GRAPHQL_REQUEST and isinstance(gql_query, GraphQLRequest):  # pyright: ignore[reportPossiblyUnboundVariable]
                        document = gql_query.document
                        variables = (
                            variable_values
                            if variable_values is not None
                            else gql_query.variable_values
                        )
                    else:
                        document = gql_query
                        variables = variable_values

                    return execute_dagster_graphql(
                        context,
                        print_ast(document),  # convert doc back to str
                        variables,
                    ).data

            with patch("dagster_graphql.client.client.Client") as mock_client:
                mock_client.return_value = MockedGraphQLClient()
                yield DagsterGraphQLClient("localhost")

    return _SpecificTestSuiteBase


ReadonlyGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_readonly_variants()
)


NonLaunchableGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_non_launchable_variants()
)

ExecutingGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_executing_variants()
)


all_repos_loadable_target = LoadableTargetOrigin(
    executable_path=sys.executable,
    python_file=file_relative_path(__file__, "cross_repo_asset_deps.py"),
)
AllRepositoryGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_executing_variants(
        target=all_repos_loadable_target, location_name="cross_asset_repos"
    )
)



from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import execute_dagster_graphql, infer_job_selector

RESOURCE_QUERY = """
query ResourceQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      modes {
        name
        resources {
          ...ResourceInfo
        }
      }
    }
  }
  resourcesOrError(pipelineSelector: $selector) {
    __typename
    ... on ResourceConnection {
      resources {
        ...ResourceInfo
      }
    }
  }
}

fragment ResourceInfo on Resource {
  name
  description
  configField {
    configType {
      key
      ... on CompositeConfigType {
        fields {
          name
          configType {
            key
          }
        }
      }
    }
  }
}
"""

REQUIRED_RESOURCE_QUERY = """
query RequiredResourceQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    ... on Pipeline {
      name
      solids {
        definition {
          ... on SolidDefinition {
            requiredResources {
              resourceKey
            }
          }
        }
      }
    }
  }
}
"""


def test_mode_fetch_resources(graphql_context: WorkspaceRequestContext, snapshot):
    selector = infer_job_selector(graphql_context, "required_resource_job")
    result = execute_dagster_graphql(
        graphql_context,
        RESOURCE_QUERY,
        {"selector": selector},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineOrError"]
    assert len(result.data["pipelineOrError"]["modes"]) == 1
    assert result.data["pipelineOrError"]["modes"][0]["resources"]
    assert (
        result.data["pipelineOrError"]["modes"][0]["resources"]
        == result.data["resourcesOrError"]["resources"]
    )

    snapshot.assert_match(result.data)


# Warning: If _compute_fields_hash changes, verify that the result.data has the same shape/keys/values
# as the existing snapshot and then run update snapshot
def test_required_resources(graphql_context: WorkspaceRequestContext, snapshot):
    selector = infer_job_selector(graphql_context, "required_resource_job")
    result = execute_dagster_graphql(
        graphql_context,
        REQUIRED_RESOURCE_QUERY,
        {"selector": selector},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineOrError"]["solids"]
    [solid] = result.data["pipelineOrError"]["solids"]
    assert solid
    assert solid["definition"]["requiredResources"]
    assert solid["definition"]["requiredResources"] == [{"resourceKey": "R1"}]

    snapshot.assert_match(result.data)



import asyncio
import logging
import os
import time

from dagster import AssetKey, Output, asset, define_asset_job
from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView, TemporalContext
from dagster._core.definitions.assets.graph.asset_graph import AssetGraph
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.execution.asset_backfill import (
    execute_asset_backfill_iteration,
    execute_asset_backfill_iteration_inner,
)
from dagster._core.execution.backfill import BulkActionStatus, PartitionBackfill
from dagster._core.execution.job_backfill import execute_job_backfill_iteration
from dagster._core.remote_origin import RemotePartitionSetOrigin
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunsFilter
from dagster._core.storage.tags import (
    ASSET_PARTITION_RANGE_END_TAG,
    ASSET_PARTITION_RANGE_START_TAG,
    BACKFILL_ID_TAG,
    PARENT_BACKFILL_ID_TAG,
    PARTITION_NAME_TAG,
    ROOT_BACKFILL_ID_TAG,
)
from dagster._core.test_utils import create_run_for_test, environ
from dagster._core.utils import make_new_backfill_id
from dagster._utils import safe_tempfile_path
from dagster_graphql.client.query import (
    LAUNCH_PARTITION_BACKFILL_MUTATION,
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
)
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    execute_dagster_graphql_and_finish_runs,
    infer_job_selector,
    infer_repository_selector,
    main_repo_location_name,
)
from dagster_shared.seven import get_system_temp_directory

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
    ReadonlyGraphQLContextTestMatrix,
)
from dagster_graphql_tests.graphql.repo import get_workspace_process_context

PARTITION_PROGRESS_QUERY = """
  query PartitionProgressQuery($backfillId: String!) {
    partitionBackfillOrError(backfillId: $backfillId) {
      ... on PartitionBackfill {
        __typename
        id
        status
        numCancelable
        partitionNames
        numPartitions
        fromFailure
        reexecutionSteps
        partitionStatuses {
          results {
            partitionName
            runStatus
          }
        }
        hasCancelPermission
        hasResumePermission
        user
        title
        description
        cancelableRuns {
          id
          runId
        }
      }
      ... on PythonError {
        message
        stack
      }
    }
  }
"""

BACKFILL_STATUS_BY_ASSET = """
  query BackfillStatusesByAsset($backfillId: String!) {
    partitionBackfillOrError(backfillId: $backfillId) {
      __typename
      ... on PartitionBackfill {
        assetBackfillData {
            assetBackfillStatuses {
                ... on AssetPartitionsStatusCounts {
                    assetKey {
                        path
                    }
                    numPartitionsTargeted
                    numPartitionsInProgress
                    numPartitionsMaterialized
                    numPartitionsFailed
                }
                ... on UnpartitionedAssetStatus {
                    assetKey {
                        path
                    }
                    inProgress
                    materialized
                    failed
                }
            }
            rootTargetedPartitions {
                partitionKeys
                ranges {
                start
                end
                }
            }
        }
      }
      ... on PythonError {
        message
        stack
      }
      ... on BackfillNotFoundError {
        message
      }
    }
  }
"""

CANCEL_BACKFILL_MUTATION = """
  mutation($backfillId: String!) {
    cancelPartitionBackfill(backfillId: $backfillId) {
      __typename
      ... on CancelBackfillSuccess {
        backfillId
      }
      ... on PythonError {
        message
        stack
      }
    }
  }
"""

RESUME_BACKFILL_MUTATION = """
  mutation($backfillId: String!) {
    resumePartitionBackfill(backfillId: $backfillId) {
      __typename
      ... on ResumeBackfillSuccess {
        backfillId
      }
      ... on PythonError {
        message
        stack
      }
    }
  }
"""

RETRY_BACKFILL_MUTATION = """
  mutation($reexecutionParams: ReexecutionParams) {
    reexecutePartitionBackfill(reexecutionParams: $reexecutionParams) {
      __typename
      ... on LaunchBackfillSuccess {
        backfillId
      }
      ... on PythonError {
        message
        stack
      }
    }
  }
"""


GET_PARTITION_BACKFILLS_QUERY = """
  query PartitionBackfillsQuery($repositorySelector: RepositorySelector!, $partitionSetName: String!) {
    partitionSetOrError(repositorySelector: $repositorySelector, partitionSetName: $partitionSetName) {
      __typename
      ...on PartitionSet {
        name
        pipelineName
        hasLaunchBackfillPermission
        hasCancelBackfillPermission
        backfills {
          id
          isAssetBackfill
        }
      }
    }
  }

"""


def _seed_runs(
    graphql_context,
    partition_runs: list[tuple[DagsterRunStatus, str | PartitionKeyRange]],
    backfill_id: str,
) -> None:
    for status, partition_or_range in partition_runs:
        partition_tags = (
            {PARTITION_NAME_TAG: partition_or_range}
            if isinstance(partition_or_range, str)
            else {
                ASSET_PARTITION_RANGE_START_TAG: partition_or_range.start,
                ASSET_PARTITION_RANGE_END_TAG: partition_or_range.end,
            }
        )
        create_run_for_test(
            instance=graphql_context.instance,
            status=status,
            tags={
                **DagsterRun.tags_for_backfill_id(backfill_id),
                **partition_tags,
            },
        )


def _get_run_stats(partition_statuses):
    return {
        "total": len(partition_statuses),
        "queued": len([status for status in partition_statuses if status["runStatus"] == "QUEUED"]),
        "in_progress": len(
            [status for status in partition_statuses if status["runStatus"] == "STARTED"]
        ),
        "success": len(
            [status for status in partition_statuses if status["runStatus"] == "SUCCESS"]
        ),
        "failure": len(
            [status for status in partition_statuses if status["runStatus"] == "FAILURE"]
        ),
        "canceled": len(
            [status for status in partition_statuses if status["runStatus"] == "CANCELED"]
        ),
    }


def _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id: str) -> None:
    """Executes an asset backfill iteration and updates the serialized asset backfill data.
    However, does not execute side effects i.e. launching runs.
    """
    backfill = graphql_context.instance.get_backfill(backfill_id)
    asset_backfill_data = backfill.asset_backfill_data
    result = None
    asset_graph_view = AssetGraphView(
        temporal_context=TemporalContext(
            effective_dt=asset_backfill_data.backfill_start_datetime, last_event_id=None
        ),
        instance=graphql_context.instance,
        asset_graph=graphql_context.asset_graph,
    )
    with environ({"ASSET_BACKFILL_CURSOR_DELAY_TIME": "0"}):
        result = execute_asset_backfill_iteration_inner(
            backfill_id=backfill_id,
            asset_backfill_data=asset_backfill_data,
            asset_graph_view=asset_graph_view,
            backfill_start_timestamp=asset_backfill_data.backfill_start_timestamp,
            logger=logging.getLogger("fake_logger"),
            run_config=None,
        )

    updated_backfill = backfill.with_asset_backfill_data(
        result.backfill_data.with_run_requests_submitted(
            result.run_requests,
            asset_graph_view=asset_graph_view,
        ),
        dynamic_partitions_store=graphql_context.instance,
        asset_graph=graphql_context.asset_graph,
    )
    graphql_context.instance.update_backfill(updated_backfill)


def _execute_backfill_iteration_with_side_effects(graphql_context, backfill_id):
    """Executes an asset backfill iteration with side effects (i.e. updates run status and bulk action status)."""
    with get_workspace_process_context(graphql_context.instance) as context:
        backfill = graphql_context.instance.get_backfill(backfill_id)
        asyncio.run(
            execute_asset_backfill_iteration(
                backfill, logging.getLogger("fake_logger"), context, graphql_context.instance
            )
        )


def _execute_job_backfill_iteration_with_side_effects(graphql_context, backfill_id):
    """Executes a job backfill iteration with side effects (i.e. updates run status and bulk action status)."""
    with get_workspace_process_context(graphql_context.instance) as context:
        backfill = graphql_context.instance.get_backfill(backfill_id)
        execute_job_backfill_iteration(
            backfill=backfill,
            logger=logging.getLogger("fake_logger"),
            workspace_process_context=context,
            instance=graphql_context.instance,
            debug_crash_flags=None,
        )


def _mock_asset_backfill_runs(
    graphql_context,
    asset_key: AssetKey,
    backfill_id: str,
    status: DagsterRunStatus,
    partition_key: str | None,
):
    partitions_def = graphql_context.asset_graph.get(asset_key).partitions_def

    @asset(
        partitions_def=partitions_def,
        name=asset_key.path[-1],
        key_prefix=asset_key.path[:-1] if len(asset_key.path) > 1 else None,
    )
    def dummy_asset():
        if status == DagsterRunStatus.FAILURE:
            raise Exception("fail")
        return Output(5)

    define_asset_job("my_job", [dummy_asset]).resolve(
        asset_graph=AssetGraph.from_assets([dummy_asset])
    ).execute_in_process(
        tags={**DagsterRun.tags_for_backfill_id(backfill_id)},
        partition_key=partition_key,
        raise_on_error=False,
        instance=graphql_context.instance,
    )


class TestPartitionBackillReadonlyFailure(ReadonlyGraphQLContextTestMatrix):
    def _create_backfill(self, graphql_context):
        code_location = graphql_context.get_code_location(main_repo_location_name())
        repository = code_location.get_repository("test_repo")

        backfill = PartitionBackfill(
            backfill_id=make_new_backfill_id(),
            partition_set_origin=RemotePartitionSetOrigin(
                repository_origin=repository.get_remote_origin(),
                partition_set_name="integer_partition",
            ),
            status=BulkActionStatus.REQUESTED,
            partition_names=["one", "two", "three"],
            from_failure=False,
            reexecution_steps=None,
            tags=None,
            backfill_timestamp=time.time(),
        )
        graphql_context.instance.add_backfill(backfill)
        return backfill.backfill_id

    def test_launch_backill_failure(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integer_partition",
                    },
                    "partitionNames": ["2", "3"],
                }
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "UnauthorizedError"

    def test_cancel_backfill_failure(self, graphql_context):
        backfill_id = self._create_backfill(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            CANCEL_BACKFILL_MUTATION,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["cancelPartitionBackfill"]["__typename"] == "UnauthorizedError"

    def test_no_permission(self, graphql_context):
        backfill_id = self._create_backfill(graphql_context)

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["hasCancelPermission"] is False
        assert result.data["partitionBackfillOrError"]["hasResumePermission"] is False

    def test_bad_id(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            BACKFILL_STATUS_BY_ASSET,
            variables={"backfillId": "Junk"},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "BackfillNotFoundError", (
            result.data
        )
        assert "Junk" in result.data["partitionBackfillOrError"]["message"]

    def test_resume_backfill_failure(self, graphql_context):
        backfill_id = self._create_backfill(graphql_context)

        result = execute_dagster_graphql(
            graphql_context,
            RESUME_BACKFILL_MUTATION,
            variables={"backfillId": backfill_id},
        )
        assert result.data
        assert result.data["resumePartitionBackfill"]["__typename"] == "UnauthorizedError", str(
            result.data
        )


class TestDaemonPartitionBackfill(ExecutingGraphQLContextTestMatrix):
    def test_launch_full_pipeline_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert result.data["partitionBackfillOrError"]["hasCancelPermission"] is True
        assert result.data["partitionBackfillOrError"]["hasResumePermission"] is True

        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2

    def test_get_partition_backfills(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        # launch a backfill for this partition set
        launch_result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                }
            },
        )
        backfill_id = launch_result.data["launchPartitionBackfill"]["backfillId"]
        result = execute_dagster_graphql(
            graphql_context,
            GET_PARTITION_BACKFILLS_QUERY,
            variables={
                "repositorySelector": repository_selector,
                "partitionSetName": "integers_partition_set",
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionSetOrError"]["__typename"] == "PartitionSet"
        assert len(result.data["partitionSetOrError"]["backfills"]) == 1
        assert result.data["partitionSetOrError"]["backfills"][0]["id"] == backfill_id
        assert result.data["partitionSetOrError"]["backfills"][0]["isAssetBackfill"] is False

    def test_launch_partial_backfill(self, graphql_context):
        # execute a full pipeline, without the failure environment variable
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        # reexecute a partial pipeline
        partial_steps = ["after_failure"]
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": partition_set_selector,
                    "partitionNames": ["2", "3"],
                    "reexecutionSteps": partial_steps,
                }
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2
        assert result.data["partitionBackfillOrError"]["reexecutionSteps"] == ["after_failure"]

    def test_cancel_job_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "hanging_partitioned_job_partition_set",
                    },
                    "partitionNames": ["1", "2"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"

        # Update backfill data to update the partition checkpoint, but manually launch the run
        # since launching the run via the backfill iteration loop will cause test process will hang forever.
        backfill = graphql_context.instance.get_backfill(backfill_id)
        partition_to_run = "1"
        graphql_context.instance.update_backfill(
            backfill.with_partition_checkpoint(partition_to_run)
        )

        # Launch the run that runs forever
        selector = infer_job_selector(graphql_context, "hanging_partitioned_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {
                            "resources": {"hanging_asset_resource": {"config": {"file": path}}}
                        },
                        "executionMetadata": {
                            "tags": [
                                {"key": "dagster/partition", "value": partition_to_run},
                                {"key": BACKFILL_ID_TAG, "value": backfill_id},
                            ]
                        },
                    }
                },
            )

            assert not result.errors
            assert result.data
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

            # ensure the execution has happened
            start = time.time()
            while not os.path.exists(path):
                time.sleep(0.1)
                assert time.time() - start < 60, "timed out waiting for file"

        runs = graphql_context.instance.get_runs(RunsFilter(tags={BACKFILL_ID_TAG: backfill_id}))
        assert len(runs) == 1
        assert runs[0].status == DagsterRunStatus.STARTED

        result = execute_dagster_graphql(
            graphql_context,
            CANCEL_BACKFILL_MUTATION,
            variables={"backfillId": backfill_id},
        )
        assert result.data
        assert result.data["cancelPartitionBackfill"]["__typename"] == "CancelBackfillSuccess"

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "CANCELING"

        start = time.time()
        while (
            graphql_context.instance.get_backfill(backfill_id).status != BulkActionStatus.CANCELED
        ):
            _execute_job_backfill_iteration_with_side_effects(graphql_context, backfill_id)
            assert time.time() - start < 60, "timed out waiting for backfill to cancel"

        runs = graphql_context.instance.get_runs(RunsFilter(tags={BACKFILL_ID_TAG: backfill_id}))
        assert len(runs) == 1
        assert runs[0].status == DagsterRunStatus.CANCELED

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "CANCELED"

    def test_cancel_then_retry_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            CANCEL_BACKFILL_MUTATION,
            variables={"backfillId": backfill_id},
        )
        assert result.data
        assert result.data["cancelPartitionBackfill"]["__typename"] == "CancelBackfillSuccess"

        # run the backfill iteration until the backfill is canceled
        while (
            graphql_context.instance.get_backfill(backfill_id).status != BulkActionStatus.CANCELED
        ):
            _execute_job_backfill_iteration_with_side_effects(graphql_context, backfill_id)

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "CANCELED"

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_backfill = graphql_context.instance.get_backfill(retried_backfill_id)

        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": retried_backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2
        assert result.data["partitionBackfillOrError"]["fromFailure"]

    def test_failing_job_backfill_cancels_runs(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "hanging_partitioned_job_partition_set",
                    },
                    "partitionNames": ["1", "2"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"

        # Update backfill data to update the partition checkpoint, but manually launch the run
        # since launching the run via the backfill iteration loop will cause test process will hang forever.
        backfill = graphql_context.instance.get_backfill(backfill_id)
        partition_to_run = "1"
        graphql_context.instance.update_backfill(
            backfill.with_partition_checkpoint(partition_to_run)
        )

        # Launch the run that runs forever
        selector = infer_job_selector(graphql_context, "hanging_partitioned_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {
                            "resources": {"hanging_asset_resource": {"config": {"file": path}}}
                        },
                        "executionMetadata": {
                            "tags": [
                                {"key": "dagster/partition", "value": partition_to_run},
                                {"key": BACKFILL_ID_TAG, "value": backfill_id},
                            ]
                        },
                    }
                },
            )

            assert not result.errors
            assert result.data
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

            # ensure the execution has happened
            start = time.time()
            while not os.path.exists(path):
                time.sleep(0.1)
                assert time.time() - start < 60, "timed out waiting for file"

        runs = graphql_context.instance.get_runs(RunsFilter(tags={BACKFILL_ID_TAG: backfill_id}))
        assert len(runs) == 1
        assert runs[0].status == DagsterRunStatus.STARTED

        # simulate the backfill failing for some reason and being marked FAILING
        updated_backfill = graphql_context.instance.get_backfill(backfill_id).with_status(
            BulkActionStatus.FAILING
        )
        graphql_context.instance.update_backfill(updated_backfill)

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "FAILING"

        start = time.time()
        while graphql_context.instance.get_backfill(backfill_id).status != BulkActionStatus.FAILED:
            _execute_job_backfill_iteration_with_side_effects(graphql_context, backfill_id)
            assert time.time() - start < 60, "timed out waiting for backfill to fail"

        runs = graphql_context.instance.get_runs(RunsFilter(tags={BACKFILL_ID_TAG: backfill_id}))
        assert len(runs) == 1
        assert runs[0].status == DagsterRunStatus.CANCELED

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "FAILED"

    def test_cancel_asset_backfill(self, graphql_context):
        asset_key = AssetKey("hanging_partition_asset")
        partitions = ["a"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input()],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        # Update asset backfill data to contain requested partition, but does not execute side effects,
        # since launching the run will cause test process will hang forever.
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # Launch the run that runs forever
        selector = infer_job_selector(graphql_context, "hanging_partition_asset_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {
                            "resources": {"hanging_asset_resource": {"config": {"file": path}}}
                        },
                        "executionMetadata": {
                            "tags": [
                                {"key": "dagster/partition", "value": "a"},
                                {"key": BACKFILL_ID_TAG, "value": backfill_id},
                            ]
                        },
                    }
                },
            )

            assert not result.errors
            assert result.data
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

            # ensure the execution has happened
            start = time.time()
            while not os.path.exists(path):
                time.sleep(0.1)
                assert time.time() - start < 60, "timed out waiting for file"

            result = execute_dagster_graphql(
                graphql_context,
                CANCEL_BACKFILL_MUTATION,
                variables={"backfillId": backfill_id},
            )
            assert result.data
            assert result.data["cancelPartitionBackfill"]["__typename"] == "CancelBackfillSuccess"

            start = time.time()
            while (
                graphql_context.instance.get_backfill(backfill_id).status
                != BulkActionStatus.CANCELED
            ):
                _execute_backfill_iteration_with_side_effects(graphql_context, backfill_id)
                assert time.time() - start < 60, "timed out waiting for backfill to cancel"

            runs = graphql_context.instance.get_runs(
                RunsFilter(tags={BACKFILL_ID_TAG: backfill_id})
            )
            assert len(runs) == 1
            assert runs[0].status == DagsterRunStatus.CANCELED

    def test_cancel_then_retry_asset_backfill(self, graphql_context):
        asset_key = AssetKey("hanging_partition_asset")
        partitions = ["a"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input()],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        # Update asset backfill data to contain requested partition, but does not execute side effects,
        # since launching the run will cause test process will hang forever.

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # Launch the run that runs forever
        selector = infer_job_selector(graphql_context, "hanging_partition_asset_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {
                            "resources": {"hanging_asset_resource": {"config": {"file": path}}}
                        },
                        "executionMetadata": {
                            "tags": [
                                {"key": "dagster/partition", "value": "a"},
                                {"key": BACKFILL_ID_TAG, "value": backfill_id},
                            ]
                        },
                    }
                },
            )

            assert not result.errors
            assert result.data
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

            # ensure the execution has happened
            start = time.time()
            while not os.path.exists(path):
                time.sleep(0.1)
                assert time.time() - start < 60, "timed out waiting for file"

            result = execute_dagster_graphql(
                graphql_context,
                CANCEL_BACKFILL_MUTATION,
                variables={"backfillId": backfill_id},
            )
            assert result.data
            assert result.data["cancelPartitionBackfill"]["__typename"] == "CancelBackfillSuccess"

            while (
                graphql_context.instance.get_backfill(backfill_id).status
                != BulkActionStatus.CANCELED
            ):
                _execute_backfill_iteration_with_side_effects(graphql_context, backfill_id)

            runs = graphql_context.instance.get_runs(
                RunsFilter(tags={BACKFILL_ID_TAG: backfill_id})
            )
            assert len(runs) == 1
            assert runs[0].status == DagsterRunStatus.CANCELED

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        # no runs were successful for the first backfill, so the retry target should be the same
        # as the first backfill
        assert (
            first_backfill.asset_backfill_data.target_subset
            == retried_backfill.asset_backfill_data.target_subset
        )
        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_failing_asset_backfill_cancels_runs(self, graphql_context):
        asset_key = AssetKey("hanging_partition_asset")
        partitions = ["a"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input()],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        # Update asset backfill data to contain requested partition, but does not execute side effects,
        # since launching the run will cause test process will hang forever.
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # Launch the run that runs forever
        selector = infer_job_selector(graphql_context, "hanging_partition_asset_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {
                            "resources": {"hanging_asset_resource": {"config": {"file": path}}}
                        },
                        "executionMetadata": {
                            "tags": [
                                {"key": "dagster/partition", "value": "a"},
                                {"key": BACKFILL_ID_TAG, "value": backfill_id},
                            ]
                        },
                    }
                },
            )

            assert not result.errors
            assert result.data
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

            # ensure the execution has happened
            start = time.time()
            while not os.path.exists(path):
                time.sleep(0.1)
                assert time.time() - start < 60, "timed out waiting for file"

            # simulate the backfill failing for some reason and being marked FAILING
            updated_backfill = graphql_context.instance.get_backfill(backfill_id).with_status(
                BulkActionStatus.FAILING
            )
            graphql_context.instance.update_backfill(updated_backfill)

            start = time.time()
            while (
                graphql_context.instance.get_backfill(backfill_id).status != BulkActionStatus.FAILED
            ):
                _execute_backfill_iteration_with_side_effects(graphql_context, backfill_id)
                assert time.time() - start < 60, "timed out waiting for backfill to fail"

            runs = graphql_context.instance.get_runs(
                RunsFilter(tags={BACKFILL_ID_TAG: backfill_id})
            )
            assert len(runs) == 1
            assert runs[0].status == DagsterRunStatus.CANCELED

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "FAILED"

    def test_resume_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2

        # manually mark as failed
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.FAILED))

        result = execute_dagster_graphql(
            graphql_context,
            RESUME_BACKFILL_MUTATION,
            variables={"backfillId": backfill_id},
        )
        assert result.data
        assert result.data["resumePartitionBackfill"]["__typename"] == "ResumeBackfillSuccess"

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"

    def test_backfill_run_stats(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3", "4", "5"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _seed_runs(
            graphql_context,
            [
                (DagsterRunStatus.SUCCESS, "5"),
                (DagsterRunStatus.STARTED, "2"),
                (DagsterRunStatus.STARTED, "3"),
                (DagsterRunStatus.STARTED, "4"),
                (DagsterRunStatus.STARTED, "5"),
                (DagsterRunStatus.CANCELED, "2"),
                (DagsterRunStatus.FAILURE, "3"),
                (DagsterRunStatus.SUCCESS, "4"),
            ],
            backfill_id,
        )

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numPartitions"] == 4
        run_stats = _get_run_stats(
            result.data["partitionBackfillOrError"]["partitionStatuses"]["results"]
        )
        assert run_stats.get("total") == 4
        assert run_stats.get("queued") == 0
        assert run_stats.get("in_progress") == 1
        assert run_stats.get("success") == 1
        assert run_stats.get("failure") == 1
        assert run_stats.get("canceled") == 1

        backfill = graphql_context.instance.get_backfill(backfill_id)

        # Artificially mark the backfill as complete - verify run status is INCOMPLETE until the runs all succeed
        graphql_context.instance.update_backfill(
            backfill.with_status(BulkActionStatus.COMPLETED_SUCCESS)
        )

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert result.data["partitionBackfillOrError"]["status"] == "COMPLETED_SUCCESS"

    def test_asset_job_backfill_run_stats(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_asset_job_partition_set",
                    },
                    "partitionNames": ["2", "3", "4", "5"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _seed_runs(
            graphql_context,
            [
                (DagsterRunStatus.SUCCESS, PartitionKeyRange("2", "5")),
            ],
            backfill_id,
        )

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numPartitions"] == 4
        run_stats = _get_run_stats(
            result.data["partitionBackfillOrError"]["partitionStatuses"]["results"]
        )
        assert run_stats.get("total") == 4
        assert run_stats.get("success") == 4

    def test_asset_backfill_stats_in_topological_order(self, graphql_context):
        asset_key_paths_in_topo_order = [
            ["upstream_static_partitioned_asset"],
            ["middle_static_partitioned_asset_1"],
            ["middle_static_partitioned_asset_2"],
            ["downstream_static_partitioned_asset"],
        ]

        partitions = ["a", "b", "c", "d", "e", "f"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [
                        AssetKey(path).to_graphql_input() for path in asset_key_paths_in_topo_order
                    ],
                }
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            BACKFILL_STATUS_BY_ASSET,
            variables={"backfillId": backfill_id},
        )

        asset_status_counts = result.data["partitionBackfillOrError"]["assetBackfillData"][
            "assetBackfillStatuses"
        ]
        assert len(asset_status_counts) == 4
        for i, path in enumerate(asset_key_paths_in_topo_order):
            assert asset_status_counts[i]["assetKey"]["path"] == path

    def test_asset_backfill_partition_stats(self, graphql_context):
        asset_key = AssetKey("upstream_static_partitioned_asset")
        partitions = ["a", "b", "c", "d", "e", "f"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input()],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        for partition, status in [
            ("a", DagsterRunStatus.SUCCESS),
            ("b", DagsterRunStatus.FAILURE),
            ("d", DagsterRunStatus.SUCCESS),
            ("e", DagsterRunStatus.SUCCESS),
            ("f", DagsterRunStatus.FAILURE),
        ]:
            _mock_asset_backfill_runs(graphql_context, asset_key, backfill_id, status, partition)

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        result = execute_dagster_graphql(
            graphql_context,
            BACKFILL_STATUS_BY_ASSET,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        backfill_data = result.data["partitionBackfillOrError"]["assetBackfillData"]

        assert backfill_data["rootTargetedPartitions"]["ranges"] is None
        assert set(backfill_data["rootTargetedPartitions"]["partitionKeys"]) == set(partitions)

        asset_partition_status_counts = backfill_data["assetBackfillStatuses"]
        assert len(asset_partition_status_counts) == 1
        assert asset_partition_status_counts[0]["assetKey"]["path"] == [
            "upstream_static_partitioned_asset"
        ]
        assert asset_partition_status_counts[0]["numPartitionsTargeted"] == 6
        assert asset_partition_status_counts[0]["numPartitionsInProgress"] == 1
        assert asset_partition_status_counts[0]["numPartitionsMaterialized"] == 3
        assert asset_partition_status_counts[0]["numPartitionsFailed"] == 2

    def test_asset_backfill_status_with_upstream_failure(self, graphql_context):
        asset_keys = [
            AssetKey("unpartitioned_upstream_of_partitioned"),
            AssetKey("upstream_daily_partitioned_asset"),
            AssetKey("downstream_weekly_partitioned_asset"),
        ]
        partitions = ["2023-01-09"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input() for asset_key in asset_keys],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("unpartitioned_upstream_of_partitioned"),
            backfill_id,
            DagsterRunStatus.SUCCESS,
            None,
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("upstream_daily_partitioned_asset"),
            backfill_id,
            DagsterRunStatus.FAILURE,
            "2023-01-09",
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        result = execute_dagster_graphql(
            graphql_context,
            BACKFILL_STATUS_BY_ASSET,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        backfill_data = result.data["partitionBackfillOrError"]["assetBackfillData"]

        assert backfill_data["rootTargetedPartitions"]["ranges"] == [
            {"start": "2023-01-09", "end": "2023-01-09"}
        ]

        asset_statuses = backfill_data["assetBackfillStatuses"]
        assert len(asset_statuses) == 3
        assert asset_statuses[0]["assetKey"]["path"] == ["unpartitioned_upstream_of_partitioned"]
        assert asset_statuses[0]["inProgress"] is False
        assert asset_statuses[0]["materialized"] is True
        assert asset_statuses[0]["failed"] is False

        assert asset_statuses[1]["assetKey"]["path"] == ["upstream_daily_partitioned_asset"]
        assert asset_statuses[1]["numPartitionsTargeted"] == 1
        assert asset_statuses[1]["numPartitionsInProgress"] == 0
        assert asset_statuses[1]["numPartitionsMaterialized"] == 0
        assert asset_statuses[1]["numPartitionsFailed"] == 1

        assert asset_statuses[2]["assetKey"]["path"] == ["downstream_weekly_partitioned_asset"]
        assert asset_statuses[2]["numPartitionsTargeted"] == 1
        assert asset_statuses[2]["numPartitionsInProgress"] == 0
        assert asset_statuses[2]["numPartitionsMaterialized"] == 0
        assert asset_statuses[2]["numPartitionsFailed"] == 1

    def test_backfill_run_completed(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3", "4", "5"],
                }
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        backfill = graphql_context.instance.get_backfill(backfill_id)

        graphql_context.instance.update_backfill(
            backfill.with_status(BulkActionStatus.COMPLETED_SUCCESS)
        )

        _seed_runs(
            graphql_context,
            [
                (DagsterRunStatus.SUCCESS, "2"),
                (DagsterRunStatus.SUCCESS, "3"),
                (DagsterRunStatus.SUCCESS, "4"),
                (DagsterRunStatus.SUCCESS, "5"),
            ],
            backfill_id,
        )

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "COMPLETED_SUCCESS"
        assert result.data["partitionBackfillOrError"]["numPartitions"] == 4

        run_stats = _get_run_stats(
            result.data["partitionBackfillOrError"]["partitionStatuses"]["results"]
        )
        assert run_stats.get("total") == 4
        assert run_stats.get("queued") == 0
        assert run_stats.get("in_progress") == 0
        assert run_stats.get("success") == 4
        assert run_stats.get("failure") == 0

    def test_backfill_run_incomplete(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3", "4", "5"],
                }
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        backfill = graphql_context.instance.get_backfill(backfill_id)

        graphql_context.instance.update_backfill(
            backfill.with_status(BulkActionStatus.COMPLETED_FAILED)
        )

        _seed_runs(
            graphql_context,
            [
                (DagsterRunStatus.SUCCESS, "2"),
                (DagsterRunStatus.SUCCESS, "3"),
                (DagsterRunStatus.STARTED, "4"),
                (DagsterRunStatus.CANCELED, "5"),
            ],
            backfill_id,
        )

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "COMPLETED_FAILED"
        assert result.data["partitionBackfillOrError"]["numPartitions"] == 4
        assert len(result.data["partitionBackfillOrError"]["cancelableRuns"]) == 1
        run_stats = _get_run_stats(
            result.data["partitionBackfillOrError"]["partitionStatuses"]["results"]
        )
        assert run_stats.get("total") == 4
        assert run_stats.get("queued") == 0
        assert run_stats.get("in_progress") == 1
        assert run_stats.get("success") == 2
        assert run_stats.get("failure") == 0
        assert run_stats.get("canceled") == 1

    def test_fetch_user_tag_from_backfill(self, graphql_context):
        user_email = "user123@abc.com"
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                    "tags": [{"key": "user", "value": user_email}],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["id"] == backfill_id
        assert result.data["partitionBackfillOrError"]["user"] == user_email

    def test_set_title_and_description_for_backfill(self, graphql_context):
        title = "Test backfill"
        description = "This is a test backfill"
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                    "title": title,
                    "description": description,
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["id"] == backfill_id
        assert result.data["partitionBackfillOrError"]["title"] == title
        assert result.data["partitionBackfillOrError"]["description"] == description

    def test_job_name_correctly_set(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        stored_backfill = graphql_context.instance.get_backfill(backfill_id)
        assert stored_backfill.job_name == "integers"

    def test_job_name_correctly_set_for_asset_backfills(self, graphql_context):
        asset_keys = [
            AssetKey("unpartitioned_upstream_of_partitioned"),
            AssetKey("upstream_daily_partitioned_asset"),
            AssetKey("downstream_weekly_partitioned_asset"),
        ]
        partitions = ["2023-01-09"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input() for asset_key in asset_keys],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        stored_backfill = graphql_context.instance.get_backfill(backfill_id)
        assert stored_backfill.job_name is None

    def test_set_title_and_description_for_backfill_invalid_title(self, graphql_context):
        title = "Title with invalid characters * %"
        description = "This is a test backfill"
        repository_selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["2", "3"],
                    "title": title,
                    "description": description,
                }
            },
        )

        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "PythonError"
        assert (
            '"Title with invalid characters * %" is not a valid title in Dagster'
            in result.data["launchPartitionBackfill"]["message"]
        )

    def test_asset_job_backfill_with_nonexistent_partition_key(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        # launch a backfill for this partition set
        launch_result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": repository_selector,
                        "partitionSetName": "integers_partition_set",
                    },
                    "partitionNames": ["1", "nonexistent1", "nonexistent2"],
                }
            },
        )
        assert (
            launch_result.data["launchPartitionBackfill"]["__typename"]
            == "PartitionKeysNotFoundError"
        )
        assert (
            "Partition keys `['nonexistent1', 'nonexistent2']` could not be found"
            in launch_result.data["launchPartitionBackfill"]["message"]
        )


class TestLaunchDaemonBackfillFromFailure(ExecutingGraphQLContextTestMatrix):
    def test_launch_from_failure(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        # trigger failure in the conditionally_fail solid

        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_pipeline_conditionally_fail"
        )
        try:
            with open(output_file, "w", encoding="utf8"):
                result = execute_dagster_graphql_and_finish_runs(
                    graphql_context,
                    LAUNCH_PARTITION_BACKFILL_MUTATION,
                    variables={
                        "backfillParams": {
                            "selector": partition_set_selector,
                            "partitionNames": ["2", "3"],
                        }
                    },
                )
        finally:
            os.remove(output_file)

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"

        # re-execute from failure (without the failure file)
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": partition_set_selector,
                    "partitionNames": ["2", "3"],
                    "fromFailure": True,
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2
        assert result.data["partitionBackfillOrError"]["fromFailure"]

    def test_launch_backfill_with_all_partitions_flag(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": partition_set_selector,
                    "allPartitions": True,
                }
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": backfill_id},
        )

        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 10
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 10

    def test_reexecute_asset_backfill_from_failure(self, graphql_context):
        asset_keys = [
            AssetKey("unpartitioned_upstream_of_partitioned"),
            AssetKey("upstream_daily_partitioned_asset"),
            AssetKey("downstream_weekly_partitioned_asset"),
        ]
        partitions = ["2023-01-09"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input() for asset_key in asset_keys],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("unpartitioned_upstream_of_partitioned"),
            backfill_id,
            DagsterRunStatus.SUCCESS,
            None,
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("upstream_daily_partitioned_asset"),
            backfill_id,
            DagsterRunStatus.FAILURE,
            "2023-01-09",
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # mark backfill as completed so we can retry it
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_from_failure_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            first_backfill.asset_backfill_data.failed_and_downstream_subset
            == retried_from_failure_backfill.asset_backfill_data.target_subset
        )
        assert retried_from_failure_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_from_failure_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            first_backfill.asset_backfill_data.target_subset
            == retried_backfill.asset_backfill_data.target_subset
        )
        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_reexecute_successful_asset_backfill(self, graphql_context):
        asset_keys = [
            AssetKey("unpartitioned_upstream_of_partitioned"),
            AssetKey("upstream_daily_partitioned_asset"),
        ]
        partitions = ["2023-01-09"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input() for asset_key in asset_keys],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("unpartitioned_upstream_of_partitioned"),
            backfill_id,
            DagsterRunStatus.SUCCESS,
            None,
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("upstream_daily_partitioned_asset"),
            backfill_id,
            DagsterRunStatus.SUCCESS,
            "2023-01-09",
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # mark backfill as complete so we can retry it
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "PythonError"
        assert (
            "Cannot re-execute from failure an asset backfill that has no missing materializations"
            in result.data["reexecutePartitionBackfill"]["message"]
        )

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"},
            },
        )
        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            first_backfill.asset_backfill_data.target_subset
            == retried_backfill.asset_backfill_data.target_subset
        )
        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_reexecute_asset_backfill_still_in_progress(self, graphql_context):
        asset_keys = [
            AssetKey("unpartitioned_upstream_of_partitioned"),
            AssetKey("upstream_daily_partitioned_asset"),
            AssetKey("downstream_weekly_partitioned_asset"),
        ]
        partitions = ["2023-01-09"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input() for asset_key in asset_keys],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("unpartitioned_upstream_of_partitioned"),
            backfill_id,
            DagsterRunStatus.SUCCESS,
            None,
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("upstream_daily_partitioned_asset"),
            backfill_id,
            DagsterRunStatus.FAILURE,
            "2023-01-09",
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # try to retry the backfill while it is still in progress
        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "PythonError"
        assert "still in progress" in result.data["reexecutePartitionBackfill"]["message"]

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "PythonError"
        assert "still in progress" in result.data["reexecutePartitionBackfill"]["message"]

        # once the first backfill is in a completed state, we can retry it
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.FAILED))

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_from_failure_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            first_backfill.asset_backfill_data.failed_and_downstream_subset
            == retried_from_failure_backfill.asset_backfill_data.target_subset
        )
        assert retried_from_failure_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_from_failure_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            first_backfill.asset_backfill_data.target_subset
            == retried_backfill.asset_backfill_data.target_subset
        )
        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_reexecute_asset_backfill_twice(self, graphql_context):
        asset_keys = [
            AssetKey("unpartitioned_upstream_of_partitioned"),
            AssetKey("upstream_daily_partitioned_asset"),
            AssetKey("downstream_weekly_partitioned_asset"),
        ]
        partitions = ["2023-01-09"]
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": partitions,
                    "assetSelection": [asset_key.to_graphql_input() for asset_key in asset_keys],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("unpartitioned_upstream_of_partitioned"),
            backfill_id,
            DagsterRunStatus.SUCCESS,
            None,
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("upstream_daily_partitioned_asset"),
            backfill_id,
            DagsterRunStatus.FAILURE,
            "2023-01-09",
        )
        _execute_asset_backfill_iteration_no_side_effects(graphql_context, backfill_id)

        # mark backfill as completed so we can retry it
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        first_backfill = graphql_context.instance.get_backfill(backfill_id)
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            first_backfill.asset_backfill_data.failed_and_downstream_subset
            == retried_backfill.asset_backfill_data.target_subset
        )
        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        _execute_asset_backfill_iteration_no_side_effects(
            graphql_context, retried_backfill.backfill_id
        )
        # mark some partitions failed so we can retry again
        _mock_asset_backfill_runs(
            graphql_context,
            AssetKey("upstream_daily_partitioned_asset"),
            retried_backfill.backfill_id,
            DagsterRunStatus.FAILURE,
            "2023-01-09",
        )
        _execute_asset_backfill_iteration_no_side_effects(
            graphql_context, retried_backfill.backfill_id
        )

        # refetch the backfill to get the updated statuses of all assets
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)
        graphql_context.instance.update_backfill(
            retried_backfill.with_status(BulkActionStatus.COMPLETED)
        )

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": retry_backfill_id, "strategy": "FROM_FAILURE"},
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        seconrd_retry_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]

        second_retried_backfill = graphql_context.instance.get_backfill(seconrd_retry_backfill_id)
        retried_backfill = graphql_context.instance.get_backfill(retry_backfill_id)

        assert (
            retried_backfill.asset_backfill_data.failed_and_downstream_subset
            == second_retried_backfill.asset_backfill_data.target_subset
        )
        assert second_retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == retry_backfill_id
        assert second_retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_retry_job_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        # trigger failure in the conditionally_fail solid

        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_pipeline_conditionally_fail"
        )
        try:
            with open(output_file, "w", encoding="utf8"):
                result = execute_dagster_graphql_and_finish_runs(
                    graphql_context,
                    LAUNCH_PARTITION_BACKFILL_MUTATION,
                    variables={
                        "backfillParams": {
                            "selector": partition_set_selector,
                            "partitionNames": ["2", "3"],
                        }
                    },
                )
        finally:
            os.remove(output_file)

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))

        # re-execute from failure (without the failure file)
        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"}
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_from_failure_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_from_failure_backfill = graphql_context.instance.get_backfill(
            retried_from_failure_backfill_id
        )

        assert retried_from_failure_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_from_failure_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": retried_from_failure_backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2
        assert result.data["partitionBackfillOrError"]["fromFailure"]

        # re-execute  (without the failure file)
        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={"reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"}},
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_backfill = graphql_context.instance.get_backfill(retried_backfill_id)

        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        result = execute_dagster_graphql(
            graphql_context,
            PARTITION_PROGRESS_QUERY,
            variables={"backfillId": retried_backfill_id},
        )
        assert not result.errors
        assert result.data
        assert result.data["partitionBackfillOrError"]["__typename"] == "PartitionBackfill"
        assert result.data["partitionBackfillOrError"]["status"] == "REQUESTED"
        assert result.data["partitionBackfillOrError"]["numCancelable"] == 2
        assert len(result.data["partitionBackfillOrError"]["partitionNames"]) == 2
        assert not result.data["partitionBackfillOrError"]["fromFailure"]

    def test_retry_in_progress_job_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        # trigger failure in the conditionally_fail solid

        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_pipeline_conditionally_fail"
        )
        try:
            with open(output_file, "w", encoding="utf8"):
                result = execute_dagster_graphql_and_finish_runs(
                    graphql_context,
                    LAUNCH_PARTITION_BACKFILL_MUTATION,
                    variables={
                        "backfillParams": {
                            "selector": partition_set_selector,
                            "partitionNames": ["2", "3"],
                        }
                    },
                )
        finally:
            os.remove(output_file)

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"}
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "PythonError"
        assert "still in progress" in result.data["reexecutePartitionBackfill"]["message"]

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={"reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"}},
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "PythonError"
        assert "still in progress" in result.data["reexecutePartitionBackfill"]["message"]

        # mark backfill as complete and confirm that we can retry it
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))

        # re-execute from failure (without the failure file)
        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"}
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_from_failure_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_from_failure_backfill = graphql_context.instance.get_backfill(
            retried_from_failure_backfill_id
        )

        assert retried_from_failure_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_from_failure_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        # re-execute (without the failure file)
        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={"reexecutionParams": {"parentRunId": backfill_id, "strategy": "ALL_STEPS"}},
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_backfill = graphql_context.instance.get_backfill(retried_backfill_id)

        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_retry_job_backfill_twice(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        # trigger failure in the conditionally_fail solid

        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_pipeline_conditionally_fail"
        )
        with open(output_file, "w", encoding="utf8"):
            result = execute_dagster_graphql_and_finish_runs(
                graphql_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "selector": partition_set_selector,
                        "partitionNames": ["2", "3"],
                    }
                },
            )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        # mark backfill as complete and confirm that we can retry it
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))
        try:
            with open(output_file, "w", encoding="utf8"):
                # re-execute from failure (still with the failure file)
                result = execute_dagster_graphql(
                    graphql_context,
                    RETRY_BACKFILL_MUTATION,
                    variables={
                        "reexecutionParams": {
                            "parentRunId": backfill_id,
                            "strategy": "FROM_FAILURE",
                        }
                    },
                )
        finally:
            os.remove(output_file)

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_backfill = graphql_context.instance.get_backfill(retried_backfill_id)

        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

        graphql_context.instance.update_backfill(
            retried_backfill.with_status(BulkActionStatus.COMPLETED)
        )

        # re-execute from failure (without the failure file)
        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {
                    "parentRunId": retried_backfill.backfill_id,
                    "strategy": "FROM_FAILURE",
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        second_retried_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        second_retried_backfill = graphql_context.instance.get_backfill(second_retried_backfill_id)

        assert second_retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == retried_backfill_id
        assert second_retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id

    def test_retry_successful_job_backfill(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        partition_set_selector = {
            "repositorySelector": repository_selector,
            "partitionSetName": "chained_failure_job_partition_set",
        }

        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": partition_set_selector,
                    "partitionNames": ["2", "3"],
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data["launchPartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        backfill_id = result.data["launchPartitionBackfill"]["backfillId"]

        # mark backfill as complete
        backfill = graphql_context.instance.get_backfill(backfill_id)
        graphql_context.instance.update_backfill(backfill.with_status(BulkActionStatus.COMPLETED))

        result = execute_dagster_graphql(
            graphql_context,
            RETRY_BACKFILL_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": backfill_id, "strategy": "FROM_FAILURE"}
            },
        )
        # Unlike asset backfills, we don't currently have a way to see if a job backfill has runs that
        # failed without querying the DB. So we always allow retries
        assert not result.errors
        assert result.data
        assert result.data["reexecutePartitionBackfill"]["__typename"] == "LaunchBackfillSuccess"
        retried_backfill_id = result.data["reexecutePartitionBackfill"]["backfillId"]
        retried_backfill = graphql_context.instance.get_backfill(retried_backfill_id)

        assert retried_backfill.tags.get(PARENT_BACKFILL_ID_TAG) == backfill_id
        assert retried_backfill.tags.get(ROOT_BACKFILL_ID_TAG) == backfill_id
